/********************************************************************
** @source
**
** @author Copyright (C) 2000,2001 Antonio Tabernero
** @version 1.30
** @last modified Ago 28 2001 Antonio Tabernero. 
** @@
**
** This program is free software; you can redistribute it and/or
** modify it under the terms of the GNU Library General Public
** License as published by the Free Software Foundation; either
** version 2 of the License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
** Library General Public License for more details.
**
** You should have received a copy of the GNU Library General Public
** License along with this library; if not, write to the
** Free Software Foundation, Inc., 59 Temple Place - Suite 330,
** Boston, MA  02111-1307, USA.
********************************************************************/
                                                                            

#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#include <string.h>

#include <time.h>
#include <sys/types.h>

#define VERSION 1.3

/**********************************************

1.3  Can use records 0x0e and 0x11 to get aprox time and position 
     instead of requiring record 0x33 (which seems to be missing 
     in some models/firmwares)

     Output Doppler data (using the flag +doppler) when applied to 
     binary files logged with the 1.2 (or newer) version of async.

************************************************/


#define AS_BYTE   0
#define AS_CHAR   1
#define AS_UINT   2
#define AS_ULONG  3
#define AS_FLOAT  4
#define AS_DOUBLE 5


#define NEVER_USED 0
#define NO_DUMP    1
#define DUMP       2
#define DUMP_PHASE_ONLY 3


#define c 2.99792458e8  // WGS84 Speed of light
#define L1 1575420000   // L1 frequency
#define lambda (c/L1)   // L1 wavelength
#define WGS84_PI 3.1415926535898


typedef unsigned char BOOLEAN;
typedef unsigned char BYTE;
typedef unsigned long ULONG;
typedef unsigned short int UINT;
typedef short int INT;


BOOLEAN ETREX; 
BOOLEAN RESET_CLOCK;
BYTE ONLY_STATS, SELECTED_SV, ONE_SAT, DIF_RECORDS, VERBOSE;
BYTE PARSE_RECORDS,RINEX_GENERATION,VERIFY_TIME_TAGS,NO_SNR;
BYTE OPT1,RELAX;

double USER_XYZ[3];
int USER_DATE[3];
BOOLEAN GIVEN_XYZ,GIVEN_DATE;


int L1_FACTOR,GET_THIS;
int SELECTED_RECORDS[16];
BYTE RINEX_FILE;
BYTE OBS_MASK,N_OBS;
UINT INTERVAL;
long START,LAST,ELAPSED;

char DATAFILE[60];
char COMMAND_LINE[256];
char location[6];
char marker[32];


////// STRUCTS  of RECORDS with (more or less) known fields

typedef struct {double llh[3]; double xyz[3];} type_rec0x11;

typedef struct {int week; int garmin_wdays; double tow;} type_rec0x0e;

typedef struct{ float delta_pr; float f1; double pr; float f2; BYTE sv; }  
type_rec0x16;

typedef struct{ BYTE sv; BYTE elev; UINT frac; UINT db; BYTE tracked; BYTE flag; }
type_rec0x1a;

typedef struct{ float altitud; float epe[3];  UINT fix; double tow; double pos[2];
                float vel[3];  float h_ellip; UINT leap; ULONG wdays;             }
type_rec0x33;


typedef struct{ ULONG c50; BYTE uk[4]; BYTE sv; }
type_rec0x36;

 
typedef struct { ULONG c_phase; 
                 long int tracked;   //BYTE  tracked; UINT cc2; BYTE flag; 
                 UINT  delta_f;  ULONG  int_phase;  
                 double pr;  
                 ULONG  c511;   
                 UINT  db;   double tow; BYTE   sv;  }  
type_rec0x38;


typedef struct{ BYTE used; UINT db; double prange; double phase; 
                float doppler; BYTE elev; long int tracked; ULONG c511; float last36;} 
rinex_obs;         


// Global variables used to study rates of change, etc.

type_rec0x38 rec38,last38;
float mean,sigma,mean_q;
float phase_rate;
float mean2=0;
float sigma2=0;
int n;
int n2=0;


void llh2xyz(double*,double*);


type_rec0x11 process_0x11(BYTE *record) // Position
{
 type_rec0x11 rec;

 memcpy(&rec.llh[0],record,8); memcpy(&rec.llh[1],record+8,8); 
 

 rec.llh[0]*=(180/WGS84_PI); rec.llh[1]*=(180/WGS84_PI); rec.llh[2]=0.0;
 llh2xyz(rec.llh,rec.xyz); 

 if (VERBOSE)
  {
   printf("Record 0x11 :: "),
   printf("Lat %.4f  Long %.4f\n",rec.llh[0],rec.llh[1]); 
   printf("XYZ(h=0.0) = [ %.0f %.0f %.0f]\n",rec.xyz[0],rec.xyz[1],rec.xyz[2]);
  }
 
 return rec;
}

double julday (year,month,day, hour, min, sec)
UINT  year,month,day,hour,min;
double sec;
{
 double jd;
 
 if (month <= 2) {year-=1; month+=12;}
 jd = floor(365.25*year)+floor(30.6001*(month+1))+day;
 jd +=(hour+(min+sec/60.0)/60.0)/24+1720981.5;
 return jd;     
}

double gps_time(double julday, int *week)
{
 double tow,days_gps;
 int wn;
 //int weekday;
 
 //weekday=floor(julday+0.5)%7  + 1 ;  // 0 Sunday, 1 Monday, ...

 // Correct at least for the 13 leap seconds as of writting this 
 julday+= (13.0/86400.0); 
 
 days_gps=(julday-2444244.5);
 wn = floor(days_gps/7);
 
 tow = (days_gps- wn*7)*86400;
 
 *week=wn;
 
 return tow;
}                             

type_rec0x0e process_0x0e(BYTE *record)   
{
 type_rec0x0e rec;
 UINT month,day,year,hour,min,sec;
 double jd;

/* 
 typedef struct {BYTE month;BYTE day;UINT year;UINT hour;BYTE min;BYTE sec;} UTC;
 UTC utc;
 utc.month=record[0];
 utc.day=record[1];
 utc.year=*((short int*)(record+2));
 utc.hour=*((short int*)(record+4));
 utc.min=record[6];
 utc.sec=record[7];
*/

 month=(UINT)record[0];
 day=(UINT)record[1];
 year=*((short int*)(record+2));
 hour=*((short int*)(record+4));
 min=(UINT)record[6];
 sec=(UINT)record[7];  
 
 jd=julday(year,month,day,hour,min,(double)sec);
 rec.tow=gps_time(jd,&rec.week);
 rec.garmin_wdays=(rec.week-521)*7;
 

  if(VERBOSE)
   { 
    printf("%02u/%02u/%4d ",day,month,year);
    printf("%02d:%02d:%02d -> JD %f -> ",hour,min,sec,jd);
    printf("GPS Week %4d ToW %6.0f sec  Garmin Weekdays %d\n",rec.week,rec.tow,rec.garmin_wdays);
   }

  return rec; 
}  

void process_0x12_gps38(BYTE *record)
{
 printf("PRN %2d ",record[3]);
 printf("TOW %14.7f ",*((double*)(record+18)));
 printf("Pseudorange %14.3f\n",*((double*)(record+26))); 
}                        




void process_0x14(BYTE *record)
{
 
 printf("Fix %d %d\n",*((UINT*)record),*((UINT*)(record+2)));
 printf("TOW4 %.7f  TOW12 %.7f\n",*((double*)(record+4)),*((double*)(record+12)));

 printf("Posicion %.4f %.4f\n",*((double*)(record+28))*180/WGS84_PI,*((double*)(record+36))*180/WGS84_PI);

 printf("Vel  %f %f %f\n",*((float*)(record+48)),*((float*)(record+52)),*((float*)(record+56)));
}                        



type_rec0x36 process_0x36(BYTE *record)
{
 type_rec0x36 rec;
 int i;

 rec.c50=*((ULONG*)(record));
 for(i=0;i<4;i++) rec.uk[i]=record[4+i];
 rec.sv=record[8];

/*
 memcpy(&rec.c50,record,4); 
 memcpy(&rec.uk,record+4,4);
 memcpy(&rec.sv,record+8,1);
 */
 if (ONE_SAT && rec.sv!=SELECTED_SV) return rec;

 if(VERBOSE)
  {
   printf("0x36 ----------------------------------------------------------------\n");
   printf("PRN %02d: C (50 Hz) %8lu = TOW %8.1f ",rec.sv+1,rec.c50,rec.c50/50.0);
   for(i=0;i<4;i++) printf("%02x ",rec.uk[i]); printf("\n");
  }

 return rec;
}


type_rec0x33 process_0x33(BYTE *record)
{
 type_rec0x33 rec;

 memcpy(&rec.altitud,record,4); record+=4;
 memcpy(&rec.epe,record,12); record+=12;
 memcpy(&rec.fix,record,2); record+=2;
 memcpy(&rec.tow,record,8); record+=8;
 memcpy(&rec.pos,record,16); record+=16;
 memcpy(&rec.vel,record,12); record+=12;
 memcpy(&rec.h_ellip,record,4); record+=4;
 memcpy(&rec.leap,record,2); record+=2;
 memcpy(&rec.wdays,record,4); record+=4;


 if(VERBOSE)
  {
 printf("0x33 ----------------------------------------------------------------\n");
 printf("Pos:  (N,E,H )  (%f %f %f)  H_ellip=%.0f  FIX %dD\n",rec.pos[0]*180/WGS84_PI,rec.pos[1]*180/WGS84_PI,rec.altitud,rec.h_ellip,rec.fix);     
 printf("Vel:  (N,E,Up)  (%f %f %f)  ",rec.vel[0],rec.vel[1],rec.vel[2]);
 printf("Epe  (dX,dY,dH) (%.1f %.1f %.1f)\n",rec.epe[0],rec.epe[1],rec.epe[2]);
 printf("Wdays %ld  TOW %13.6f  Leap %d\n",rec.wdays,rec.tow,rec.leap);
  }

 return rec;
}


void process_0x37(BYTE *record)
{
 double pr,tow;
 ULONG c511;
 BYTE svid,unknown[4];
 UINT cc,countdown,c1,c2;
 float TR;

 memcpy(&c1,record,2);
 memcpy(unknown,record+2,2); 
 memcpy(&c2,record+4,2);
 memcpy(&cc,record+6,2);
 memcpy(&countdown,record+8,2); TR=((float)1024.0-countdown)/16.0;
 memcpy(&pr,record+10,8);
 memcpy(&c511,record+18,4);
 memcpy(&tow,record+22,8);
 memcpy(unknown+2,record+30,2);
 svid=record[32];


 if (ONE_SAT && svid!=SELECTED_SV) return;

 printf("0x37 ----------------------------------------------------------------\n");
 printf("PRN %02d: PR %10.1f TOW %10.3f (c511 %10lu) CC %5d\n",svid+1,pr,tow,c511,cc);
 printf("       Countdown %5d (%2.0f'') C1 %04x C2 %04x\n",countdown,TR,c1,c2);
 printf("       UNKNOWN: ");
 printf("%02x %02x | %02x %02x\n",unknown[0],unknown[1],unknown[2],unknown[3]);  
}

void get_0x1a_info(BYTE *record, type_rec0x1a chan[])
{
 BYTE *ptr;
 int k;

 for(k=0,ptr=record;k<12;k++,ptr+=8)
  {
  if (ETREX)
   {
   chan[k].sv=ptr[5];
   chan[k].elev=ptr[4];
   chan[k].frac=*((UINT*)(ptr+0));
   chan[k].db=*((UINT*)(ptr+2));
   chan[k].tracked=ptr[6];
   chan[k].flag=ptr[7];
   }
  else
   {
   chan[k].sv=ptr[0];
   chan[k].elev=ptr[1];
   chan[k].frac=*((UINT*)(ptr+2));
   chan[k].db=*((UINT*)(ptr+4));
   chan[k].tracked=ptr[6];
   chan[k].flag=ptr[7];
   }
  }                

 return;
}

void process_0x1a(BYTE *record)
{
 int k;
 BYTE *rec;
 BYTE svid,elev,trk,flag;
 unsigned int sq;
 float frac_phase;
 
 
 if (!ONE_SAT)
 {
  printf("0x1a ---------------------------------------------------------------\n");
  printf("Chann -> PRN | Elev | Q Sgnl |  Phase  | Track | Flag |\n");
 }
 
 for(k=0,rec=record;k<12;k++,rec+=8)
 {
  if (ETREX)
   {
    svid=rec[5]; elev=rec[4]; sq=*((UINT*)(rec+2));
    frac_phase=*((UINT*)(rec+0))/2048.0; trk=rec[6]; flag=rec[7];
   }
  else
   {
    svid=rec[0]; elev=rec[1]; sq=*((UINT*)(rec+4));
    frac_phase=*((UINT*)(rec+2))/2048.0; trk=rec[6]; flag=rec[7];
   }                        
  if (ONE_SAT)
  {
   if (svid!=SELECTED_SV) continue;
   printf("0x1a ----------------------------------------------------------------\n");
   printf("PRN %02d: Chann %02d  Elev %2d  Q %5d ",svid+1,k,elev,sq);
   printf("( Phase ) %5.3f  Trk %d Flag %d\n",frac_phase,trk,flag);
  }
  else
  {
   printf("  %02d  -> ",k);
   if (svid==0xff) printf("not used\n");
   else
   {
    printf("%02d  |  %02d  | %6d |",svid+1,elev,sq);
    printf("  %05.3f |  %2u  |  %2u  |\n",frac_phase,trk,flag);
   }
  }
 }
 //printf("----------------------------------------------------------\n");
 //getch();
}                

type_rec0x16 process_0x16(BYTE *record)
{
 type_rec0x16 rec;
 float dif;
 
 rec.sv=record[20];
 if ( ONE_SAT && rec.sv != SELECTED_SV) return rec;
 
 if (ETREX)
  {
   memcpy(&rec.pr,record,8);
   memcpy(&rec.delta_pr,record+8,4);
   memcpy(&rec.f1,record+12,4);
   memcpy(&rec.f2,record+16,4);
  }
 else
  {
   memcpy(&rec.delta_pr,record,4);
   memcpy(&rec.f1,record+4,4);
   memcpy(&rec.pr,record+8,8);
   memcpy(&rec.f2,record+16,4);
  }
                            
if (VERBOSE)
 {              
  printf("0x16 ---------------------------------------------------------\n");
  printf("PRN %02d: ",rec.sv+1);
  printf("Pseudorange %13.3f ",rec.pr);
  printf("Doppler(m/s)  %7.2f -> %.3f Hz ",rec.delta_pr,rec.delta_pr/lambda);
  printf("F1 %.2f F2 %.2f\n",rec.f1,rec.f2);
  if (ETREX) printf("Unknown %d %d %d\n",record[21],record[22],record[23]);
 
  if ( (DIF_RECORDS) && (ONE_SAT) && (rec.pr==rec38.pr) )
   {
    dif=(rec.delta_pr-phase_rate)*100;
    if (fabs(dif)<5000)
     {
      mean2+=dif; sigma2+=(dif*dif); n2++;
      printf("\t(delta_pr-phase_rate) -> ");
      printf("Mean %.1f std %.1f\n",mean2/n2,sqrt((sigma2-mean2*mean2/n2)/n2));
     }
   }
 }

 return rec; 
}
                    
BYTE process_0x39(BYTE *record)
{
 BYTE sv;

 sv=record[34];
 if (ONE_SAT && (sv != SELECTED_SV))  return sv;    

 mean=sigma=0.0; n=0;

 if (VERBOSE)
  {
   printf("0x39 ---------------------------------------------------------\n");    
   printf("PRN %02d: ACQUIRED PSEUDORANGE\n",record[34]+1);
  }

 return sv;
}


ULONG tobin(BYTE b)
{
 ULONG res;
 BYTE mask=128;
 int k;

 res=0;
 for(k=0,mask=128;k<8;k++,mask/=2) 
  {
   res=(res*10)+( (b&mask)>>(7-k)) ;
  }
  
 return res;
}

type_rec0x38 process_0x38(BYTE* record)
{
 int nsec,doppler;
 double dt,slip,d_pr,drift,d_phase,phase;
 ULONG d_cphase,d_intphase,d_c511;
 type_rec0x38 rec;
 long dtracked;
 long itow,dtow; 
 
 last38=rec38;
 
 if (ONE_SAT && ( record[36] != SELECTED_SV)) {rec.sv=record[36]; return rec;}
 
 if (ETREX)
  {
   memcpy(&(rec38.pr),record,8);
   memcpy(&(rec38.tow),record+8,8);
   memcpy(&(rec38.c_phase),record+16,4);
   memcpy(&(rec38.tracked),record+20,4);
   memcpy(&(rec38.int_phase),record+24,4);
   memcpy(&(rec38.c511),record+28,4);
   memcpy(&(rec38.delta_f),record+32,2);
   memcpy(&(rec38.db),record+34,2);
   memcpy(&(rec38.sv),record+36,1);
  }
 else
  {
   memcpy(&(rec38.c_phase),record,4);
   memcpy(&(rec38.tracked),record+4,4);
   memcpy(&(rec38.delta_f),record+8,2);
   memcpy(&(rec38.int_phase),record+10,4);
   memcpy(&(rec38.pr),record+14,8);
   memcpy(&(rec38.c511),record+22,4);
 
   memcpy(&(rec38.db),record+26,2);
   memcpy(&(rec38.tow),record+28,8);
   memcpy(&(rec38.sv),record+36,1);
  }
                                   
 //rec38.delta_f=*((UINT*)(record+8));
 //rec38.int_phase=*((ULONG*)(record+10));
 //rec38.pr=*((double*)(record+14));
 //rec38.c511=*((ULONG*)(record+22));
 //rec38.db=*((UINT*)(record+26));
 //rec38.tow=*((double*)(record+28));
 //rec38.sv=record[36];
 
 doppler=rec38.delta_f-32768;
 phase=rec38.int_phase+(rec38.c_phase & 2047)/2048.0;
 

 if(VERBOSE)
  {
  // Check that tow is within limits
  itow=(long)floor(rec38.tow+0.5);  if(START==-1) START=itow;
  dtow=itow-START;
  if( (itow>=START) && (itow<=LAST) && (dtow<ELAPSED) ) 
   {
    printf("0x38 ----------------------------------------------------------------\n");
    printf("PRN %02d: Sgn Q %5d ",rec38.sv+1,rec38.db);
    printf("TRACKED %08x %010d -> %.3f\n",rec38.tracked,(int)rec38.tracked,(float)rec38.tracked/256.0);
    printf("\tTOW %14.7f (Counter 0.5 Mhz = %8lu).\n",rec38.tow,rec38.c511);
    printf("\tPseudoRange  %14.3f  Integrated Phase %14.3f\n",rec38.pr,phase);
 
    printf("\tDoppler(\?): %5d Hz -> %.2f m/s\n",rec38.delta_f,rec38.delta_f*lambda);
   }
 }
                                       
 if (ONE_SAT && DIF_RECORDS)   // Differencing records is only valid when tracking one sat
  {
   dt=rec38.tow-last38.tow;
   nsec=floor(dt+0.499999999);                  // GPS time exact increment
   drift = (dt-nsec)/nsec;                      // Clock drift from the 1'' increment
   d_pr=(rec38.pr-last38.pr);                   // Pseudorange increment
   d_cphase=rec38.c_phase-last38.c_phase;       // Phase counter increment
   d_phase=d_cphase/2048.0;                     // Phase increment
   d_intphase=(rec38.int_phase-last38.int_phase);  // Integer phase increment   phase_rate=d_phase*lambda/nsec;        // phase rate (in cm/s)
   d_c511=rec38.c511-last38.c511;
   slip= (d_pr - d_phase*lambda)*100;   // discrepancy between phase and pseudorange
   if (fabs(slip)<5000)
    {
     mean_q+=rec38.db;
     mean+=(slip/nsec); sigma+=(slip*slip/(nsec*nsec)); n++;
    }
   dtracked=rec38.tracked-last38.tracked;
 
   if(VERBOSE)
    {
   printf("Deltas:");
   printf(" 0.5Mhz %7ld  Clk drift: %.1f musec/sec -> (m/s) %.1f\n",d_c511,drift*1e6,drift*c);
   printf("\tdTracked %8d\n",dtracked);
   printf("\tdPR(m) %.1f dintPhase %u  dPhase %.3f dPR(m/s) %.1f\n",d_pr,d_intphase,d_phase,d_pr/dt);
   printf("\td_PR-d_PHASE %.0f cm ",slip);
   printf("(mean %.1f, std %.1f) ",mean/n,sqrt((sigma-mean*mean/n)/n));
   printf("Mean Q %.0f\n",mean_q/n);
    }
  }
 
 return rec38;
}                     



void collect_stats(FILE *org)
{
 BYTE id,L,record[256];
 int k;
 BYTE lengths[256];
 ULONG cont[256];
 BYTE var[256];

 for(k=0;k<256;k++) cont[k]=var[k]=0;

 do 
 {
  fread(&id,1,1,org); fread(&L,1,1,org); fread(record,1,L,org); //fseek(org,L,SEEK_CUR);    

  if ((cont[id])  && (lengths[id]!=L) ) var[id]=1;
  cont[id]++; 
  lengths[id]=L;
 }
 while (feof(org)==0);
 fclose(org);

 
 for (k=0;k<256;k++)
 {
  if (cont[k]==0) continue;
  printf("Record 0x%02x  (%5ld) L=%3d bytes ",k,cont[k],lengths[k]);
  if (var[k]) printf("(VAR)");
  printf("\n");
 }

  exit(0);
  return;
}


void check(BYTE *record,BYTE L,BYTE type)
{
 int k,size;


 switch(type)
 {
  case AS_BYTE: for(k=0;k<L;k++) printf("%02x ",record[k]); printf("\n"); break;
  case AS_CHAR: for(k=0;k<L;k++) printf("%c",record[k]); printf("\n"); break;
  case AS_DOUBLE: size=sizeof(double); if(L<size) return;
          for(k=0;k<=L-size;k++) printf("%3d %f\n",k,*((double*)(record+k)));
          break;
  case AS_FLOAT: size=sizeof(float); if(L<size) return;
          for(k=0;k<=L-size;k++) printf("%3d %f\n",k,*((float*)(record+k)));
          break;  
  case AS_ULONG: size=sizeof(ULONG); if(L<size) return;
          for(k=0;k<=L-size;k++) printf("%3d %lu\n",k,*((ULONG*)(record+k)));
          break;  
  case AS_UINT: size=sizeof(UINT); if(L<size) return;    
          for(k=0;k<=L-size;k++) printf("%3d %u\n",k,*((UINT*)(record+k)));
          break;
  default: break;
 }
 printf("------------------------------------------------------------------------------\n"); 

}


int found_in_list(int lista[], BYTE id)
{
 int k=0;
 int found=0;

 while (lista[k]!=-1)
 {
  if (lista[k]==id) found=1; k++;
 }      

 return found;
}


void reset_epoch(rinex_obs list[])
{
 int k;

 for(k=0;k<32;k++) 
  {
   if (list[k].used!=NEVER_USED) list[k].used=NO_DUMP; 
   list[k].doppler=-1;
  } 

}
 
BYTE get_q_code(UINT sgn)
{
 //float temp;
 //temp=(float)sgn/2000.0 + 3.0; if (temp>9) temp=9;
 //return (BYTE)floor(temp);

 if (sgn>=9000) return 9;
 if (sgn>=6000) return 8;
 if (sgn>=4000) return 7;
 if (sgn>=2500) return 6;
 return 5;
}                         



UINT get_uint(BYTE* ptr)
  {
   UINT x;
 
   x= (UINT)(*ptr++);
   x+=256*(UINT)(*ptr);
 
   return x;
  }    

void get_prod_id(org,description,prod,version)
FILE *org;
char *description;
UINT *prod;
float *version;
{
 BYTE id,L,record[256];
 UINT temp,k;
 BOOLEAN bad;
 
 do
 {
  fread(&id,1,1,org); fread(&L,1,1,org); fread(record,1,L,org);
  switch (id)
   {
    case 0xff: 
      bad=0;k=4;   // Check if it is a good looking ID record 
      while(record[k++] && !bad) bad=(record[k]>=128);
      if(!bad) // Probably  a good ff record
       {
        temp=get_uint(record); *prod=temp; 
        temp=get_uint(record+2); *version=(float)temp/100;
        strcpy(description,record+4);
        rewind(org);
        return;
       }
      else break;
    default  : break;
   }
 }
 while (feof(org)==0);

 *prod=0; *version=0.0; strcpy(description,"Generic GPS12");
 rewind(org);
 return; 
} 


void llh2xyz(double llh[],double xyz[])
{
 double a,f,a_WGS84,f_WGS84;
 double e2,N,temp;
 double lat,lon,h;
 double ellip[2]={0,0};
 //double ellip[2]={-251.0, -0.14192702};  Hayford (European)

 a_WGS84=6378137; f_WGS84=1/298.257223563;
 a = a_WGS84 - ellip[0];
 f = f_WGS84 - ellip[1]*1e-4;
 
 e2=f*(2-f);
 
 lat=WGS84_PI*llh[0]/180;
 lon=WGS84_PI*llh[1]/180;
 h=llh[2];
                    
 N=a/sqrt(1-e2*sin(lat)*sin(lat));
 
 temp=(N+h)*cos(lat);

 xyz[0]=temp*cos(lon);
 xyz[1]=temp*sin(lon);
 xyz[2]=(N*(1-e2)+h)*sin(lat);
}


void get_wdays_and_tow_from_user_date(ULONG* wdays,ULONG* week_secs)   
{
 UINT month,day,year,hour,min,sec;
 int gps_week;
 double jd,tow;

 year=(UINT)USER_DATE[0];
 month=(UINT)USER_DATE[1];
 day=(UINT)USER_DATE[2];
 hour=(UINT)12; min=(UINT)0; sec=(UINT)0;  
 
 jd=julday(year,month,day,hour,min,(double)sec);
 tow=(ULONG)gps_time(jd,&gps_week); *week_secs=(ULONG)tow;
 *wdays=(gps_week-521)*7;
 

 if(VERBOSE)
 { 
   printf("%02u/%02u/%4d ",day,month,year);
   printf("%02d:%02d:%02d -> JD %f :: ",hour,min,sec,jd);
   printf("GPS Week %4d ToW %6.0f sec  Garmin Weekdays %d\n",gps_week,tow,*wdays);
 }
 
}  


void get_aprox_location_and_wdays_and_tow(org,xyz,wdays,tow)
FILE *org;
double xyz[3];
ULONG *wdays;
ULONG *tow;
{
 BYTE id,L,record[256];
 type_rec0x33 rec;
 type_rec0x11 rec11;
 type_rec0x0e rec0e;
 int fix=0;
 int k=0;
 BYTE found_11,found_0e,found_33;

 found_33=found_11=found_0e=0;
 
 
 do
 {
  fread(&id,1,1,org); fread(&L,1,1,org); fread(record,1,L,org);
 
  switch (id)  
   {
    case 0x33: 
               rec=process_0x33(record); k++; if(k<GET_THIS) break;
               fix=rec.fix;

               xyz[0]=180.0*rec.pos[0]/WGS84_PI;
               xyz[1]=180.0*rec.pos[1]/WGS84_PI;
               xyz[2]=rec.altitud-rec.h_ellip;
               llh2xyz(xyz,xyz);
			   
               *wdays=rec.wdays;  
               *tow=(ULONG)floor(rec.tow+0.5);
	
			   found_33=1; break;

    case 0x0e: rec0e=process_0x0e(record); found_0e=1; break;

    case 0x11: rec11=process_0x11(record); found_11=1; break;

    default  : break;
   }
 }
 while (feof(org)==0);


 rewind(org);


 if(found_33)   // Found 0x33 record
  {
   //printf("Obtained date and position from 0x33 record.Fix = %d.\n",fix);
   if (fix<3)
    {
     printf("Couldn't detect a 3D fix in this session\n");
     printf("You probably won't get a useful RINEX file, but let's try it\n");
    }
   //return;
  }    
 else 
  {
   if (found_0e && found_11)   // Found 0x11 AND 0x0e records
    {
     //printf("Using 0x0e and 0x11 records to get date & position\n");
     for(k=0;k<3;k++) xyz[k]=rec11.llh[k]; llh2xyz(xyz,xyz);

     *wdays=rec0e.garmin_wdays;
     *tow=(ULONG)rec0e.tow;
     //printf("[%f %f %f] :: %d %u\n",pos[0],pos[1],pos[2],*wdays,*tow);
    }
   else         // No pertinent records found 
    {
     if ( (GIVEN_DATE==0) || (GIVEN_XYZ==0) )
	 {		 
      printf("Couldn't find any records to get aproximate location and date\n");
      printf("Make sure that you are using the latest version of async to log your data\n");
      printf("You can also overcome this problem by providing the date of observation\n");
      printf("(using the -date DD MM YYYY option) AND your approximate position\n");
	  printf("(using the -llh or -xyz options) in the command line\n");
	  printf("-------------------------------------------------------------------------\n");
	  printf("Exiting now\n");
      exit(0);
	 } 
    } 
  }



 // If position or date were provided in the command line 
 // they are always used with preference to those found 

 if (GIVEN_XYZ) for(k=0;k<3;k++) xyz[k]=USER_XYZ[k];
 
 if (GIVEN_DATE) get_wdays_and_tow_from_user_date(wdays,tow);


}




/////////////////////////////////////////////////7

void get_current_date(ULONG wdays,double tow,struct tm *gmt)
{
 time_t week_start,current;
 ULONG seconds;
 struct tm *qq;

 ULONG TIME_START=631065600L;  // Difference beetween GPS time and UNIX time          

 seconds=(ULONG)floor(tow);
 
 week_start=TIME_START+wdays*24*3600L;
 current=week_start+seconds;
 qq=gmtime(&current); 
 *gmt=*qq;

}

char *padd(char *in, char *out, int maxL)
{
 size_t L,cop;
 int k;

 L=strlen(in);
 cop = (L>maxL)? maxL:L;

 for(k=0;k<cop;k++) out[k]=in[k];
 for(k=cop;k<maxL;k++) out[k]=' ';
 out[maxL]=0;

 return out;
}



void generate_rinex_header(xyz,wdays,first_obs,prod_id,version,description,fd)
double xyz[],first_obs;
float version;
ULONG wdays;
UINT prod_id;
char *description;
FILE *fd;
{
 BYTE mask;
 int k,l,lines,written,max_lines,nchars;
 char *header,*ptr;
 time_t tt;
 char *date,buffer[80];
 struct tm gmt;
 double dt,secs;
 char obs[3][3]= {"C1", "L1", "D1"};
 

 // I dont think I'll use more than 40 lines for the header

 max_lines=40;
 header=(char*)malloc(max_lines*80*sizeof(char));  

 ptr=header;
 
 written=sprintf(ptr,"%6d%14c",2,32); ptr+=written;
 written=sprintf(ptr,padd("OBSERVATION FILE",buffer,20)); ptr+=written;
 written=sprintf(ptr,padd("GPS NAVSTAR",buffer,20)); ptr+=written;
 written=sprintf(ptr,padd("RINEX VERSION / TYPE",buffer,20)); ptr+=written;

 sprintf(buffer,"GAR2RNX %4.2f",VERSION);
 written=sprintf(ptr,padd(buffer,buffer,20)); ptr+=written; 
 written=sprintf(ptr,padd("Any GPS12 Owner",buffer,20)); ptr+=written; 
 time(&tt); date=ctime(&tt); 
 written=sprintf(ptr,padd(date,buffer,20)); ptr+=written;
 written=sprintf(ptr,"PGM / RUN BY / DATE "); ptr+=written; 
 
 padd("** gar2rnx (Garmin to Rinex) generates Rinex2 files",buffer,60);
 written=sprintf(ptr,buffer);  ptr+=written; 
 written=sprintf(ptr,padd("COMMENT",buffer,20)); ptr+=written;

 padd("** from a GPS12 (or XL) (Copyright Antonio Tabernero)",buffer,60);
 written=sprintf(ptr,"%s",buffer);            ptr+=written;
 written=sprintf(ptr,"COMMENT             "); ptr+=written;


 sprintf(buffer,"** Generated from G12 data file: %s ",DATAFILE);
 padd(buffer,buffer,60); 
 written=sprintf(ptr,"%s",buffer); ptr+=written;
 written=sprintf(ptr,"%s",padd("COMMENT",buffer,20)); ptr+=written;

 written=sprintf(ptr,"** Options: "); ptr+=written;
 written=sprintf(ptr,padd(COMMAND_LINE,buffer,60-written)); ptr+=written;
 written=sprintf(ptr,"COMMENT             "); ptr+=written;  

 sprintf(buffer,"%s",marker);
 written=sprintf(ptr,padd(buffer,buffer,60));   ptr+=written; 
 written=sprintf(ptr,padd("MARKER NAME",buffer,20));              ptr+=written;
 written=sprintf(ptr,padd("Number of Antenna marker",buffer,60)); ptr+=written;
 written=sprintf(ptr,padd("MARKER NUMBER",buffer,20));            ptr+=written;
 written=sprintf(ptr,padd("Your name here",buffer,20));           ptr+=written;
 written=sprintf(ptr,padd("Your boss here",buffer,40));           ptr+=written;
 written=sprintf(ptr,"OBSERVER / AGENCY   "); ptr+=written;

 written=sprintf(ptr,"%03d%17c",prod_id,32); 
 if (written>20) written=20; ptr+=written; 
 written=sprintf(ptr,"%s  ",padd(description,buffer,18)); ptr+=20;
 written=sprintf(ptr,"%4.2f%16c",version,32); 
 if (written>20) written=20; ptr+=written;   
 written=sprintf(ptr,"REC # / TYPE / VERS "); ptr+=written;
 

 written=sprintf(ptr,"%d%19c",0,32); ptr+=written;
 written=sprintf(ptr,padd("GPS12 Internal Antenna",buffer,40)); ptr+=written; 
 written=sprintf(ptr,padd("ANT # / TYPE",buffer,20));           ptr+=written; 

 if (GIVEN_XYZ) padd("** Position provided by user in command line",buffer,60);
 else padd("** Position from first 3D fix of the receiver",buffer,60);
 written=sprintf(ptr,buffer); ptr+=written;
 written=sprintf(ptr,"COMMENT             "); ptr+=written;

 written=sprintf(ptr,"%14.4f%14.4f%14.4f%18c",xyz[0],xyz[1],xyz[2],32); 
 ptr+=written;
 written=sprintf(ptr,"APPROX POSITION XYZ "); ptr+=written; 

 written=sprintf(ptr,"%14.4f%14.4f%14.4f%18c",0.0,0.0,0.0,32); ptr+=written;
 written=sprintf(ptr,"ANTENNA: DELTA H/E/N"); ptr+=written;


 written=sprintf(ptr,"%6d%6d%48cWAVELENGTH FACT L1/2",L1_FACTOR,0,32);  
 ptr+=written; 


 
 written=sprintf(ptr,"%6d",N_OBS); ptr+=written;

 for(k=0,mask=1;k<3;k++,mask*=2) 
	{
     if (OBS_MASK & mask) {written=sprintf(ptr,"%6s",obs[k]);  ptr+=written;} 
	}
 for(k=N_OBS;k<9;k++) {written=sprintf(ptr,"%6c",32);  ptr+=written; }
 written=sprintf(ptr,"# / TYPES OF OBSERV "); ptr+=written; 


 if (RESET_CLOCK) 
  { dt=first_obs-floor(first_obs+0.5); dt=floor(dt*1e9)/1e9; first_obs-=dt; }

 get_current_date(wdays,first_obs,&gmt);
 written=sprintf(ptr,"%6d%6d",1900+gmt.tm_year,gmt.tm_mon+1); ptr+=written; 
 written=sprintf(ptr,"%6d%6d%6d",gmt.tm_mday,gmt.tm_hour,gmt.tm_min); 
 ptr+=written; 
 secs=(gmt.tm_sec)+first_obs-floor(first_obs);
 written=sprintf(ptr,"%12.6f%6cGPS%9c",secs,32,32); ptr+=written; 
 written=sprintf(ptr,"TIME OF FIRST OBS   "); ptr+=written;   

 written=sprintf(ptr,"%6d%54cINTERVAL%12c",INTERVAL,32,32);  ptr+=written; 
 written=sprintf(ptr,"%60cEND OF HEADER       ",32);  ptr+=written;


 nchars=(ptr-header);
 lines=nchars/80;
 // printf("Number of chars %d -> lines %d\n",nchars,lines);

 for(l=0;l<lines;l++,fprintf(fd,"\n")) 
   for(k=0;k<80;k++)  fprintf(fd,"%c",header[l*80+k]);

 free((char*)header);                                                                                
}              



////////////////////////////////////////////////////


void print_rinex_info(ULONG wdays, double tow,rinex_obs epoch[],FILE *fd)
{
 int k,N_used;
 double frac,phase,pr;
 int snr;
 struct tm gmt;

 double dt;
 
 // printf("TOW %.0f Start %d Last %d\n",tow,START,START+ELAPSED);

 for(k=0,N_used=0;k<32;k++) if (epoch[k].used>=DUMP) N_used++;
 
 if (N_used)  
  {
   if (RESET_CLOCK) { dt=tow-floor(tow+0.5); dt=floor(dt*1e9)/1e9;} else dt=0.0;

   get_current_date(wdays,tow-dt,&gmt);
   frac=(tow-dt)-floor(tow-dt);
   fprintf(fd," %02d %02d",(1900+gmt.tm_year)%100,gmt.tm_mon+1);
   fprintf(fd," %02d %02d %02d",gmt.tm_mday,gmt.tm_hour,gmt.tm_min); 
   fprintf(fd,"%11.7f",(double)gmt.tm_sec+frac);
   //fprintf(fd,"%11.7f",(double)gmt.tm_sec+frac,32,32);
   
   fprintf(fd,"%3d%3d",0,N_used);
   for(k=0;k<32;k++) if(epoch[k].used>=DUMP) fprintf(fd,"G%2d",k+1); 
   for(k=0;k<12-N_used;k++) fprintf(fd,"%3c",32);

   //if(RESET_CLOCK) fprintf(fd,"%12.9f",dt); 
   fprintf(fd,"\n");

   for(k=0;k<32;k++) if(epoch[k].used>=DUMP)
    {
     snr= (NO_SNR)? 0:get_q_code(epoch[k].db);    
	  
     pr=epoch[k].prange-c*dt; 
     phase=epoch[k].phase-L1*dt;


	 /*
     if(N_OBS>=1)
      {  
       if(epoch[k].used==DUMP_PHASE_ONLY) fprintf(fd,"%16c",32); 
       else fprintf(fd,"%14.3f%1c%1d",pr,32,snr); 
      }
     if(N_OBS>=2)  fprintf(fd,"%14.3f%1c%1d",phase,32,snr);

     if( (N_OBS>=3) && (epoch[k].doppler!=-1))  
        fprintf(fd,"%14.3f%1c%1d",-epoch[k].doppler,32,snr); 
*/
	 
     if (OBS_MASK&1) // Pseudoranges
	 {
      if(epoch[k].used==DUMP_PHASE_ONLY) fprintf(fd,"%16c",32); 
      else fprintf(fd,"%14.3f%1c%1d",pr,32,snr); 
	 }

     if (OBS_MASK&2)  // L1 Phase
      fprintf(fd,"%14.3f%1c%1d",phase,32,snr);


     if ( (OBS_MASK&4) &&  (epoch[k].doppler!=-1))  //Doppler
      fprintf(fd,"%14.3f%1c%1d",-epoch[k].doppler,32,snr); 


     fprintf(fd,"\n"); 
      
    }

  }

}              


void add_0x1a_to_epoch(rinex_obs epoch[], type_rec0x1a chan[])
{
 int k;
 BYTE sv;
 
 for(k=0;k<12;k++)
  {
   sv=chan[k].sv; if (sv>=32) continue;
   if (epoch[sv].db==chan[k].db)
    epoch[sv].elev=chan[k].elev;  
  }
}        




void get_session_number(char *name, char *ext)
{
 int k;
 FILE *fd;
 char fich[64];

 for(k=1;k<10;k++)
  {
   sprintf(fich,"%s%1d.%s%c",name,k,ext,0);
   fd=fopen(fich,"r");
   if (fd==NULL) break;
   fclose(fd);
  }

 strcpy(name,fich); 
 
}
 
void get_rinex_file_name(station,week_days,week_secs,filename)
char *station,*filename;
ULONG week_days,week_secs;
{
 struct tm date;
 char ext[8];
 
 get_current_date(week_days,(double)week_secs,&date);
 sprintf(filename,"%4s%03d%c",station,date.tm_yday+1,0);
 sprintf(ext,"%02dO%c",(1900+date.tm_year)%100,0);
 get_session_number(filename,ext);

}
 



void verify_tt(FILE *org)
{
 ULONG check[32][2],cc;
 BYTE flag[32],ff;


 float tow[32],current_tow;
 //BYTE flag38[32];

 BYTE id,L,record[256],sv;
 int k;
 type_rec0x38 rec;
 type_rec0x36 rec36;



 for(k=0;k<32;k++) {flag[k]=0xff; tow[k]=-1;}

 do
 {
  fread(&id,1,1,org); fread(&L,1,1,org); fread(record,1,L,org);   

  switch (id)
   {
    case 0x36: 
      rec36=process_0x36(record); sv=rec36.sv; if(sv>32) break;
      if(ONE_SAT && (sv!=SELECTED_SV) ) break;

      cc=rec36.c50; ff=flag[sv];     
      if (ff==0xff) {flag[sv]=0; check[sv][1]=cc; break;}

      if (cc==(check[sv][1]+30))  // OK
       {
        if (ff==1)    // print report
         {
          printf("PRN %02d: 0x36 gap ",sv+1);
          printf("(%.1f -> %.1f)\n",check[sv][0]/50.0,check[sv][1]/50.0);
          flag[sv]=0;
         }
        check[sv][0]=check[sv][1]; 
       }
      else  flag[sv]=1;       // Lost 0x36 record
      check[sv][1]=cc; 

      break;

    case 0x38: 
      rec=process_0x38(record);  sv=rec.sv; 

      if(sv>32) break;  
      if(ONE_SAT && (sv!=SELECTED_SV) ) break;
 
      current_tow=(float)floor(rec.tow+0.5); 
      
      //printf("%.0f %.0f\n",tow[sv],current_tow);

      if (tow[sv]==-1) {tow[sv]=current_tow; break;}
          
 
      if (current_tow != (tow[sv]+1)  )  
       {
        printf("PRN %02d: ",sv+1); 
        if (current_tow<tow[sv]) 
          printf("Anomalous tow %.0f -> %.0f\n",tow[sv],current_tow);
        else if (current_tow==tow[sv]) printf("%.0f Duplicado\n",current_tow);
             else
              {
               printf("Missing %.0f",tow[sv]+1);
               if (current_tow == (tow[sv]+2) ) printf("\n");
               else printf(" -> %.0f)\n",current_tow-1);
               tow[sv]=current_tow; 
              }
       }
      else tow[sv]=current_tow;           
      
      break;
 
    default  : break;
   }
 }
 while (feof(org)==0);
 
 fclose(org);
 exit(0);
}       

type_rec0x38 choose(type_rec0x38 rec[],int nn, rinex_obs last)
{
 int k;
 type_rec0x38 best;
 float dif[2],DIF,dif_min;
 long int dt;

 // Only one case
 if (nn==1) { best=rec[0]; return best; }

 // No previous references (any will do)
 if (last.used==NEVER_USED)  best=rec[nn-1]; 
 else  // Choose between several
   {
    dif_min=(float)1e30;
    for(k=0;k<nn;k++)
     {
      dt=rec[k].tracked-last.tracked;
      dif[0]=(float)dt/256.0;
      dif[1]=(float)rec[k].delta_f-last.doppler; 
      DIF=fabs(dif[0])+fabs(dif[1]);
      if (DIF<dif_min) {dif_min=DIF; best=rec[k];}
     }   
   }
 return best; 
}

int verify_c511(type_rec0x38 allrec[],int N,double last_tow,ULONG next_c511)
 {
  int k,j,nsat,nmax;
  int n_511,nn[40],test;
  ULONG c511[40],cc,c511_ok;
  float dif;

  n_511=0; for(j=0;j<40;j++) nn[j]=0;

  for (k=0;k<N;k++)
   {
    cc=allrec[k].c511; test=0;
    for (j=0;j<n_511;j++) if (cc==c511[j]) { nn[j]++; test=1; break;}      
    if (test==0)  { c511[n_511]=cc; nn[n_511]=1; n_511++; }
   } 

  nmax=0;
  for(k=0;k<n_511;k++) if (nn[k]>nmax) { c511_ok=c511[k]; nmax=nn[k]; }
  if (nmax==0) return 0;

  dif=(float)next_c511-(float)c511_ok;
  if (last_tow!=-1) if (abs(dif)>20) return 0; 

  nsat=0;
  for(k=0;k<N;k++) if (allrec[k].c511==c511_ok) allrec[nsat++]=allrec[k];
   

  //printf("Verifying c511: %2d records: Expecting %12u\n",N,next_c511);
  //for(k=0;k<n_511;k++) printf("C511   %16u -> %2d veces\n",c511[k],nn[k]);
  //                     printf("Chosen %16u con %2d veces Dif %.0f\n",c511_ok,nmax,dif);

  return nsat;
 }

int remove_duplicates(type_rec0x38 allrec[],int N,rinex_obs epoch[])
 {
  int k,j,nsat,sv,used[32],nn;
  type_rec0x38 rec[20];
  double tow,phase;
  
  tow=allrec[0].tow;

  // Find SVs in sight and possibly duplicated records

  for(k=0;k<32;k++) used[k]=0;
  for (k=0;k<N;k++) used[allrec[k].sv]++; 

  //printf("Remove: Tow %14.6f -> ",tow);
  //printf("NREC %2d :: ",N); for(k=0;k<N;k++) printf("%2d ",allrec[k].sv+1); 
  //printf("\n");  


  nsat=0;
  for(k=0;k<N;k++)
   {
    nn=0;
    rec[nn++]=allrec[k]; sv=rec[0].sv; 
    if (used[sv]<0) continue;        // This sv has already been processed 

    if (used[sv]>1) 
     {
      for(j=k+1;j<N;j++) 
       { if (allrec[j].sv==sv) rec[nn++]=allrec[j]; if (nn==used[sv]) break; } 
      used[sv]=-1;
     }
    allrec[nsat++]=choose(rec,nn,epoch[sv]); 
   } 

  return nsat;

  //printf("N SV %2d :: ",nsat); for(k=0;k<nsat;k++) printf("%2d ",allrec[k].sv+1);
  //printf("\n");

  for(k=0;k<nsat;k++) 
  {
    if ((allrec[k].sv+1)!=21) continue;
    phase=allrec[k].int_phase+(allrec[k].c_phase & 2047)/2048.0;
    printf("0x38 ----------------------------------------------------------------\n");
    printf("PRN %02d: Sgn Q %5d ",allrec[k].sv+1,allrec[k].db);
    printf("TRACKED %08x -> %8d\n",allrec[k].tracked,(int)allrec[k].tracked);
    printf("\tTOW %12.5f (Counter 0.5 Mhz = %u).\n",allrec[k].tow,allrec[k].c511);
    printf("\tPseudoRange  %14.3f  Integrated Phase %14.3f\n",allrec[k].pr,phase);
    printf("\tDoppler(Hz): %5d \n",allrec[k].delta_f);
  }

  return nsat;
 }



void add_0x16_to_epoch(rinex_obs epoch[], type_rec0x16 rec)
{
 BYTE sv=rec.sv;

 //printf("%d %14.3f %14.3f %f\n",sv,epoch[sv].prange,rec.pr,epoch[sv].prange-rec.pr);
 
 if (epoch[sv].prange==rec.pr) epoch[sv].doppler=rec.delta_pr/lambda;

}

void add_0x38_to_epoch(rinex_obs epoch[], type_rec0x38 rec)
{
 BYTE sv;
 int action;
 BYTE check;
 long  dtracked;

 sv=rec.sv;

 check=rec.tracked&0xff;

 if (epoch[sv].used==NEVER_USED)  action= (check==0)? NO_DUMP:DUMP;
 else
  {
   dtracked=(epoch[sv].tracked-rec.tracked);
   action = (abs(dtracked)<256)?  DUMP:NO_DUMP;
  }

 if (RELAX) action=DUMP;

 
 // Check if there have been a recent 0x36 record for that sat

 if( (rec.tow>(epoch[sv].last36+2.0)) && (action==DUMP) )  
   action = (OPT1) ? DUMP_PHASE_ONLY:NO_DUMP;
  
 //printf("SV %2d -> %08x %08x : ",sv+1,epoch[sv].tracked,rec.tracked,dtracked);
 //printf("dTracked %8d :",dtracked);
 //printf("check %02x action %d\n",check,action);

 // Check for meaningless values
 if ( (rec.pr<=10.0) || (rec.pr>=1e9) ) return;
 //if (rec.db>15000)  return;  

 //epoch[sv].doppler=rec.delta_f; 
 epoch[sv].prange=rec.pr;
 epoch[sv].phase=(double)rec.int_phase+(rec.c_phase&2047)/2048.0;
 epoch[sv].db=rec.db;
 
 epoch[sv].tracked=rec.tracked;
 epoch[sv].c511=rec.c511;

 epoch[sv].used=action;

}


int process_tow(rec,N,rec16,N16,epoch,last_tow)
type_rec0x38 *rec;
int N;
type_rec0x16 *rec16;
int N16;
rinex_obs *epoch;
double last_tow;
{
 double current_tow=rec[0].tow;
 int n_sat;
 int k;


 // Check for current_tow < last_tow (problems during week rollover)
 //if (current_tow<last_tow) return 0;   

 // Check for tow=0.000000
 if (current_tow==0.0) return 0;


 //Eliminates duplicated records 
 n_sat=remove_duplicates(rec,N,epoch);


 //printf("Tow %14.6f -> %2d records -> %2d sats\n",current_tow,N,n_sat); 

 // Add (or not) records to rinex obs.
 for(k=0;k<n_sat;k++) add_0x38_to_epoch(epoch,rec[k]);

 // Add doppler data
 for(k=0;k<N16;k++) add_0x16_to_epoch(epoch,rec16[k]);

 return n_sat;
}






void generate_rinex(FILE *org)
{
 BYTE id,L,record[256],sv;
 int k,n_records,nr,n_16;
 type_rec0x16 r16,rec16[48];
 type_rec0x38 rec,allrec[48];
 type_rec0x36 rec36;
 //type_rec0x1a chan[12];
 double current_tow,last_tow;
 ULONG next_c511,last_c511=0;
 rinex_obs epoch[32];
 ULONG week_days,week_secs;
 double aprox_xyz[3];
 char description[256];
 UINT prod_number;
 long itow,dtow; 
 float version;

 BOOLEAN found=0;
 long fptr;
 
 FILE *dest;
 char name[128];


 get_prod_id(org,description,&prod_number,&version); 
 get_aprox_location_and_wdays_and_tow(org,aprox_xyz,&week_days,&week_secs);


 if (RINEX_FILE) 
  {
   get_rinex_file_name(location,week_days,week_secs,name);  
   dest=fopen(name,"w");
  }
 else dest=stdout;


 //printf("Week days %d TOW %d -> File %s\n",week_days,week_secs,name); 
 //printf("Aprox location %f %f %f\n",aprox_llh[0],aprox_llh[1],aprox_llh[2]);
 //printf("Aprox XYZ  %f %f %f\n",aprox_xyz[0],aprox_xyz[1],aprox_xyz[2]);
 //printf("ID %d.\n Desc: %s .\n Soft %.2f\n",prod_number,description,version);
 //exit(0);

 rewind(org);

 n_records=0; n_16=0;
 last_tow=-1;

 reset_epoch(epoch);
 for(k=0;k<32;k++) {epoch[k].used=NEVER_USED; epoch[k].last36=-1.0;}

 do
 {
  fptr=ftell(org);
  fread(&id,1,1,org); fread(&L,1,1,org); fread(record,1,L,org); 
  fptr-=ftell(org); 

  switch (id)
   {
    case 0x1a:   
      break;   
   
    case 0x16:
   
      r16=process_0x16(record); sv=r16.sv; if(sv>=32) break;


	  // Option A: keep all of them and discard them later
      if (n_16<48) rec16[n_16++]=r16;
      //if (n_16>=12) {printf("N16 %d Tow %.0f\n",n_16,current_tow); getchar();}


	  // OPtion B: check to see if there is already a 0x16 record for that sv
	  //found=0; for (k=0;k<n_16;k++) if (sv==rec16[k].sv) {found=1; break;}
	  //if (found==0) rec16[n_16++]=r16;
	  //else { printf("Repeated 0x16 for SV %d, tow %.0f. Not added\n",sv,current_tow);}

       
      break;
 
    case 0x36:   
      rec36=process_0x36(record); sv=rec36.sv; if(sv>=32) break;
      epoch[sv].last36=(float)rec36.c50/50.0;
      break;

    case 0x38: 
      
      rec=process_0x38(record);  sv=rec.sv; if(sv>=32) break;  

      if (n_records==0) current_tow=rec.tow; 

      if (rec.tow==current_tow) allrec[n_records++]=rec; 
      else                             // End of epoch detected
       {
        fseek(org,fptr,SEEK_CUR);      // Reset file pointer for next epoch
        nr=n_records; n_records=0;

        //printf("End TOW:  0x38 records: %d  ",nr); 

        // Check that tow is within limits
        itow=(long)floor(current_tow+0.5);  if(START==-1) START=itow;
        dtow=itow-START;
        if( (itow<START) || (itow>LAST) || (dtow>ELAPSED) )  {n_16=0; break;}
     
        // Verifies c511 fields
       
        //printf("%10.3f (%2d) ->\n ",current_tow,nr); 

        next_c511 = last_c511 + (ULONG)floor((current_tow-last_tow)*511500.0 +0.5);
        nr=verify_c511(allrec,nr,last_tow,next_c511); if (nr==0) {n_16=0; break;}

        //printf("After c511 check = %2d. 0x16 records %d\n",nr,n_16); 

        nr=process_tow(allrec,nr,rec16,n_16,epoch,last_tow);


        //printf("Final %2d:  ",nr); 
        //for(k=0;k<nr;k++) printf("%02d ",allrec[k].sv+1); printf("\n"); 
 
        if (nr==0) {n_16=0; break;}
 

        // If first epoch, creates header 
        if (last_tow==-1)          
          generate_rinex_header(aprox_xyz,week_days,current_tow,\
                                prod_number,version,description,dest);

        // If multiple of interval, dump to rinex file 
        if( (INTERVAL==1) || (itow%INTERVAL)==0 )
         print_rinex_info(week_days,current_tow,epoch,dest);                     
       
        //for(k=0;k<n_16;k++) printf("%02d %14.3f\n",rec16[k].sv+1,rec16[k].pr); 
        //getchar();

        n_16=0;  // Reset number of 0x16 records per epoch

        last_c511=allrec[0].c511;
        
        //printf("Expected c511 %u -> seen %u\n",next_c511,last_c511);
        last_tow=current_tow;
        reset_epoch(epoch);
       }
      
      break;
 
    default  : break;
   }
 }
 while (feof(org)==0);
 
 fclose(org);
 if (RINEX_FILE) fclose(dest);

 exit(0);
}       



                  




void parse_records(FILE *org)
{
 BYTE record[256],id,L;

 fread(&id,1,1,org); fread(&L,1,1,org); fread(record,1,L,org);   
 
 if (!found_in_list(SELECTED_RECORDS,id)) return;

 switch (id)
  {
   case 0x12: process_0x12_gps38(record); break;

   case 0x0e: process_0x0e(record);  break;        
   case 0x11: process_0x11(record);  break;        
   case 0x14: process_0x14(record);  break;        
   case 0x16: process_0x16(record);  break;    
   case 0x1a: process_0x1a(record);  break;	  
   case 0x33: process_0x33(record);  break;
   case 0x36: process_0x36(record);  break;	  
   case 0x37: process_0x37(record);  break;
   case 0x38: process_0x38(record);  break;     
   case 0x39: process_0x39(record);  break;     
   default :
     printf("Record %02x ------------------------------------------\n",id);  
     check(record,L,AS_BYTE); getchar();
     break;  
 }

}


void original_parsing(FILE *fd)
{
  do parse_records(fd); while (feof(fd)==0);

  if (DIF_RECORDS) 
   {
     printf("%2d    %5.0f        ",SELECTED_SV+1,(float)mean_q/n);
     printf("%4.1f  %5.1f   ",mean/n,sqrt((sigma-mean*mean/n)/n));
     printf("%4.1f  %5.1f\n",mean2/n2,sqrt((sigma2-mean2*mean2/n2)/n2));
   }
  fclose(fd);
  exit(0);
}


void print_help(char **argv)
{
 char help[16384];

  sprintf(help,"-----------------------------------------------------------------\n\
 Gar2rnx (Garmin to Rinex) generates rinex2 compliant files from \n\
 data obtained with some Garmin handhelds using the async logger \n\
-----------------------------------------------------------------\n\
* Version %4.2f    Copyright 2000,2001  Antonio Tabernero Galan *\n\
-----------------------------------------------------------------\n",VERSION);


  strcat(help,"\nUsage: gar2rnx g12file [-stat] [-parse options] [-rinex options]\n\
\n\
  g12file is a file generated using the async logger utility.\n\
          It's the only mandatory argument. On this data file,\n\
          the gar2rnx program can perform three basic functions,\n\
          and those can have its own options\n\
\n\
------------------------------------------------------------------\n\
  -stat : shows statistics about the number, Identity byte and\n\
          length of received packets\n\
\n");

 
 strcat(help,"------------------------------------------------------------------\n\
  -parse: this was the original functionality, designed to help us\n\
          interprete the async messages sent by the GPS12.\n\
\n\
   List of options that can be activated along with -parse:\n\
\n\
  -sats: shows records 0x1a (Track status, signal, elevation, etc.)\n\
         Default\n\
  -all : displays info about all sats. Default.\n\
  -s nn: for those fields with a SV ID field, displays only\n\
         those records regarding the satellite with PRN nn\n\
  -r 0xr1 0xr2 ... : displays info about records 0xr1 0xr2 (hexadecimal)\n\
         The fields (more or less) known are interpreted.\n\
         Unknown bytes are simply listed.\n\
  -dif : presents some additional info about the increment of several\n\
         fields in records 0x38 and 0x16 (Obsolete)\n\
\n");

 
 strcat(help,"------------------------------------------------------------------\n\
  -rinex: this is the default functionality (hence the change of\n\
          name of the program from parser to gar2rnx).\n\
          Since it's the default you actually don't have to type it.\n\
\n\
   List of options that can be used along with -rinex:\n\
\n\
  -etrex    : when using an Etrex (or Emap) to log your data, you\n\
               should use this option to obtain a proper Rinex file.\n\
  -reset    : reset the time-tags to the nearest full second\n\
               modifying the observables accordingly\n\
  -f        : Instead of sending the RINEX file to standard output\n\
               (default) it creates a file using the RINEX conventions\n\
\n------------------------------------------------------------------\n\n\
   -start tow: starts the generation of the RINEX file from tow\n\
               (week_seconds). By default, it starts from the first\n\
               record logged in g12file\n\
   -stop  tow: don't include in the RINEX file those records with a\n\
               time-tag later than tow (week_seconds)\n\
   -time  tt : only tt seconds of observations go into the RINEX file\n\
   -int   tt : only those records from g12file that are time-tagged\n\
               with a multiple of tt seconds go into the RINEX file\n\
\n------------------------------------------------------------------\n");

 strcat(help,"\n\
   Observables dumped into the RINEX file:\n\
   By default, gar2rnx will output pseudoranges and L1 phases.\n\
   These observables can be modified using the following options:\n\
\n\
   -code -phase -doppler: eliminate the corresponding observables\n\
                          from the default set (code+phase)\n\
   +code +phase +doppler: add  the corresponding observables\n\
                          to the default set (code+phase)\n\
   For instance: -phase   : will output only pseudoranges\n\
                 +doppler : will dump code, phase and doppler\n\
                 -code +doppler: will output phase and doppler\n");


strcat(help,"------------------------------------------------------------------\n\
   -date YYYY MM DD: date of the first observation (year month day)\n\
                     You should only use this option if you get\n\
                     an error message telling you that the date\n\
                     couldn't be obtained from the binary file.\n\
\n\
   -xyz X Y Z : our approximate position in ECEF XYZ coordinates (meters).\n\
   -llh lat long height: same but using lat-long (deg) and height (mt) (WGS84)\n\
\n\
   There are two situations where the above options are to be used:\n\
     * When no proper position could be obtained from the binary file.\n\
       Wait until you see an error message telling you so when running gar2rnx\n\
     * When you want to use that rinex file as data from a reference station\n\
       and you know the coordinates of your observing site with\n\
       a better precision than a navigation solution (+/- 10 mt)\n");
  

 
strcat(help,"\n---------------------------------------------------------------\n\
   -area name: Name of the station or area where observations are taken.\n\
               If the option -f is given, the first 4 chars of this name\n\
               are used to create the RINEX filename.\n\
\n\
   -mark name: Marker name in the Rinex file. Name used to identify\n\
               a particular point of measurement.\n\
\n\
------------------------------------------------------------------\n");

 printf("%s",help); 
 exit(0);   


}


FILE* parse_arg(int argc, char **argv)
{
 int arg_num,j;
 FILE *fd;
 //BOOLEAN first_obs;

 if (argc<2) print_help(argv);

 strcpy(DATAFILE,argv[1]);
 if (argc>2) 
  {
   strcpy(COMMAND_LINE,argv[2]);
   for(j=3;j<argc;j++) {strcat(COMMAND_LINE," "); strcat(COMMAND_LINE,argv[j]);}
  }
   
 fd=fopen(argv[1],"rb");
 if (fd==NULL ) { printf("Cannot open binary file %s (\?\')\n",argv[1]); exit(0);}
 
 // Defaults values: shows record 1a (sat info) for all satellites.

 SELECTED_SV=0x14;
 ONE_SAT=0;
 SELECTED_RECORDS[0]=0x1a; SELECTED_RECORDS[1]=-1;
 DIF_RECORDS=0;

 ETREX=0;
 RINEX_GENERATION=1;
 VERBOSE=0;

 ONLY_STATS=0;
 PARSE_RECORDS=0;
 VERIFY_TIME_TAGS=0;
 NO_SNR=0;


 OBS_MASK=3;     // Dump code + phase
 RINEX_FILE=0;   // Send Rinex to stdout
 L1_FACTOR=1;
 GET_THIS=1;

 INTERVAL=1;
 START=-1;
 LAST=604800+1;
 ELAPSED=604800+1;

 RELAX=0;
 OPT1=0;

 GIVEN_XYZ=0; GIVEN_DATE=0;


 strcpy(location,"area");
 strcpy(marker,"Garmin");

 // User provided arguments
 arg_num=2;
 while (arg_num<argc)
  {	
   if (strcmp(argv[arg_num],"-stat")==0) { ONLY_STATS=1; arg_num++;}    
   else if (strcmp(argv[arg_num],"-all")==0) { ONE_SAT=0; arg_num++;}     
   else if (strcmp(argv[arg_num],"-dif")==0) { DIF_RECORDS=1; arg_num++;}     
   else if (strcmp(argv[arg_num],"-reset")==0) { RESET_CLOCK=1; arg_num++;}    
   else if (strcmp(argv[arg_num],"-etrex")==0) { ETREX=1; arg_num++;}  
   else if (strcmp(argv[arg_num],"-parse")==0) 
    { PARSE_RECORDS=1; RINEX_GENERATION=0; VERBOSE=1; arg_num++;}     
   else if (strcmp(argv[arg_num],"-o1")==0) 
    { OPT1=1; arg_num++; }    
   else if (strcmp(argv[arg_num],"-o2")==0) 
    { RELAX=1; arg_num++; }    
   else if (strcmp(argv[arg_num],"-verify")==0) 
    { VERIFY_TIME_TAGS=1; RINEX_GENERATION=0; arg_num++; }    
   else if (strcmp(argv[arg_num],"-rinex")==0) 
    { RINEX_GENERATION=1; VERBOSE=0; arg_num++;}     
   else if (strcmp(argv[arg_num],"-half")==0) 
    { L1_FACTOR=2; arg_num++;}     
   else if (strcmp(argv[arg_num],"-f")==0) 
    { RINEX_FILE=1; arg_num++; }      
   else if (strcmp(argv[arg_num],"-snroff")==0) { NO_SNR=1; arg_num++;}     
   else if (strcmp(argv[arg_num],"-get33")==0) 
    { GET_THIS=atoi(argv[arg_num+1]); arg_num+=2; }      
   else if (strcmp(argv[arg_num],"-start")==0) 
    { START=atoi(argv[arg_num+1]); arg_num+=2; }                            
   else if (strcmp(argv[arg_num],"-stop")==0) 
    { LAST=atoi(argv[arg_num+1]); arg_num+=2; }                            
   else if (strcmp(argv[arg_num],"-time")==0) 
    { ELAPSED=atoi(argv[arg_num+1]); arg_num+=2; }                            
   else if (strcmp(argv[arg_num],"-int")==0) 
    { INTERVAL=atoi(argv[arg_num+1]); arg_num+=2; }                            
   //else if (strcmp(argv[arg_num],"-obs")==0) { N_OBS=atoi(argv[arg_num+1]); arg_num+=2; }                            
   else if (strcmp(argv[arg_num],"-code")==0) 
	{
	 OBS_MASK = OBS_MASK & 6; arg_num++;
	}  
   else if (strcmp(argv[arg_num],"-phase")==0) 
	{
	 OBS_MASK = OBS_MASK & 5; arg_num++;
	}
   else if (strcmp(argv[arg_num],"-doppler")==0) 
	{
	 OBS_MASK = OBS_MASK & 3; arg_num++;
	}
   else if (strcmp(argv[arg_num],"+code")==0) 
	{
	 OBS_MASK = OBS_MASK | 1; arg_num++;
	}   
    else if (strcmp(argv[arg_num],"+phase")==0) 
	{
	 OBS_MASK = OBS_MASK | 2; arg_num++;
	}      
    else if (strcmp(argv[arg_num],"+doppler")==0) 
	{
	 OBS_MASK = OBS_MASK | 4; arg_num++;
	}   
   else if (strcmp(argv[arg_num],"-area")==0) 
    { strncpy(location,argv[arg_num+1],4); arg_num+=2; }                            
   else if (strcmp(argv[arg_num],"-mark")==0) 
    { strcpy(marker,argv[arg_num+1]); arg_num+=2; }                            
   else if (strcmp(argv[arg_num],"-sats")==0) 
    { 
     ONE_SAT=0; SELECTED_RECORDS[0]=0x1a;  SELECTED_RECORDS[1]=-1; arg_num++; 
    }    
   else if (strcmp(argv[arg_num],"-s")==0) 
    { 
     ONE_SAT=1; SELECTED_SV=atoi(argv[arg_num+1])-1; arg_num+=2;
    }                            
   else if (strcmp(argv[arg_num],"-r")==0)
    {
     arg_num++; j=0;
     while ((arg_num!=argc) && (argv[arg_num][0]!=45) )
       SELECTED_RECORDS[j++]=strtoul(argv[arg_num++],(char**)NULL,16); 		
     SELECTED_RECORDS[j]=-1;
    }
   else if (strcmp(argv[arg_num],"-date")==0) 
    { 
     GIVEN_DATE=1; arg_num++;
	 for (j=0;j<3;j++)
	 {	       
      if  ( (arg_num==argc) || (argv[arg_num][0]==45) )
	  {
       printf("Correct usage of -date option: -date YYYY MM DD (date of observation)\n"); exit(0);
	  }
	  else USER_DATE[j]=atoi(argv[arg_num++]); 
	 }
    }  
   else if (strcmp(argv[arg_num],"-xyz")==0) 
    { 
     GIVEN_XYZ=1; arg_num++;
	 for (j=0;j<3;j++)
	 {	       
      if  ( (arg_num==argc) )
	  {
       printf("Correct usage of -xyz option: -xyz X Y Z (ECEF position in meters)\n"); exit(0);
	  }
	  else USER_XYZ[j]=atof(argv[arg_num++]); 
	 }
    }
   else if (strcmp(argv[arg_num],"-llh")==0) 
    { 
     GIVEN_XYZ=1; arg_num++;
	 for (j=0;j<3;j++)
	 {	       
      if  ( (arg_num==argc) )
	  {
       printf("Correct usage of -llh option: -llh lat long h [deg deg mt])\n"); exit(0);
	  }
	  else USER_XYZ[j]=atof(argv[arg_num++]); 
	 }
     llh2xyz(USER_XYZ,USER_XYZ);
    }  
   else { printf("Unknown Option %s\n",argv[arg_num]); arg_num++;}	
  }

  if (VERBOSE)
  {
   if (GIVEN_DATE) printf("DATE: %2d/%2d/%4d\n",USER_DATE[2],USER_DATE[1],USER_DATE[0]);
   if (GIVEN_XYZ) printf("XYZ : %f %f %f\n",USER_XYZ[0],USER_XYZ[1],USER_XYZ[2]);  
  }


  N_OBS=(OBS_MASK & 1) + ((OBS_MASK & 2)>>1) + ((OBS_MASK & 4)>>2);
  if (N_OBS==0) 
	{
	 printf("You have selected no observables for the RINEX file\n");
	 printf("Exiting now\n");
	 exit(0);
	}


  // If we just want code it doesnt make sense to allow phase_only output
  if ( OBS_MASK==1) OPT1=0; 


  //printf("Mask %d. N_obs %d. Observables: ",OBS_MASK,N_OBS);
  //if (OBS_MASK & 1) printf("code ");
  //if (OBS_MASK & 2) printf("phase ");
  //if (OBS_MASK & 4) printf("doppler ");
  //printf("\n");
  //exit(0);

   return fd;
}


int main(int argc,char**argv)
{
  FILE *fd;
  


  fd=parse_arg(argc,argv);

  if (ONLY_STATS) collect_stats(fd);
  if (PARSE_RECORDS) original_parsing(fd);
  if (VERIFY_TIME_TAGS) verify_tt(fd);   
  generate_rinex(fd);

  return 0;
}


